/*
 * メインルーチン
 */

#include "common.h"

extern "C"
{
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <termios.h>
#include <stdarg.h>
#include <time.h>
#include <ctype.h>
#include <locale.h>
#include <sys/types.h>
#include <pwd.h>
#include <dirent.h>

#ifdef __CYGWIN__
#include <ncurses/term.h>
#include <sys/time.h>
#else
#include <term.h>
#endif

}

///////////////////////////////////////////////////////////////////////////////
// 内部で使う変数
///////////////////////////////////////////////////////////////////////////////
const int kKeyMetaFirst = 128;      // Metaキーコードの最初
const int kKeyEsc = 27;             // Escキーコード

static char gHostName[256];         // ホスト名
static char gUserName[256];         // ユーザー名

///////////////////////////////////////////////////////////////////////////////
// 外にさらす変数
///////////////////////////////////////////////////////////////////////////////
cDirWnd* gLDir;                         // 左のディレクトリオブジェクト
cDirWnd* gRDir;                         // 右のディレクトリオブジェクト
bool gMainLoop = false;                 // メインループ
hash_obj* gKeyCommand[2][KEY_MAX];      // キーとコマンドのマップ[0][キーコード] metaなし [1][キーコード] meta
char gTrashBoxPath[PATH_MAX];           // ごみ箱のディレクトリパス

bool gExtensionICase = true;            // マークファイルの拡張子別実行で大文字小文字を区別するかどうか
bool gColor = false;                    // 画面の描写にカラーを使うかどうか
bool gIndividualCursor = true;          // 左右のファイラ画面でカーソル位置が独立しているかどうか
bool gCheckDeleteFile = true;           // ファイルを消す時に確認するかどうか
bool gCheckCopyFile = false;            // ファイルを消す時に確認するかどうか
bool gCheckExit = true;                 // ファイラを終了する時に確認する稼動
bool gGnuScreen = false;                // GNU screen上で起動しているかどうか
bool gAutoRehash = false;               // コマンドラインオートリハッシュ
bool gUnixDomainSocket = true;          // UnixDomainSocketを使うかどうか
bool gRemainCursor = false;             // カーソル位置を保存するかどうか
bool gExecutiveFileBold = true;
bool gReadDirHistory = true;
enum eKanjiCode gKanjiCodeFileName = kUnknown;     // ファイル名の漢字コード kUnknownの場合は自動判別

bool gBoldExe = true;
bool gBoldDir = true;
int gColorMark = kCAYellow;
int gColorDir = kCACyan;
int gColorExe = kCARed;
int gColorLink = kCAMagenta;

char* gVersion = "3.1.8";

void (*gViewSystem)() = NULL;      // 登録描写関数
void (*gView)() = NULL;  // ユーザー用登録描写関数

bool gChangeTerminalTitle = true;

static void initialize_global_vars(char* init_path)
{
    char* home = getenv("HOME");
    if(home == NULL) {
        fprintf(stderr, "$HOME is NULL\n");
        exit(1);
    }

    strcpy(gTrashBoxPath, home);
    strcat(gTrashBoxPath, "/mtrashbox");
    
    for(int i=0; i<2; i++) {
        for(int j=0; j<KEY_MAX; j++) {
            gKeyCommand[i][j] = hash_new(10);
         }
    }

    gethostname(gHostName, 256);
    struct passwd* user = getpwuid(getuid());
    strcpy(gUserName, user->pw_name);

    if(strcmp(init_path, "") == 0) {  
        gLDir = new cDirWnd("", false);
        gRDir = new cDirWnd("", true);
    }
    else {
        struct stat _stat;
        if(stat(init_path, &_stat) < 0) {
            fprintf(stderr, "%s don't exist\n", init_path);
            exit(1);
        }
        if(!S_ISDIR(_stat.st_mode)) {
            fprintf(stderr, "%s is not directory\n", init_path);
            exit(1);
        }
    
        gLDir = new cDirWnd(init_path, false);
        gRDir = new cDirWnd(init_path, true);
    }
}

static void finalize_global_vars()
{
    delete gLDir;
    delete gRDir;

    for(int i=0; i<2; i++) {
        for(int j=0; j<KEY_MAX; j++) {
            for(hash_it* it = gKeyCommand[i][j]->mEntryIt;
                it;
                it = it->mNextIt)
            {
                FREE(it->mItem);
            }
            hash_delete(gKeyCommand[i][j]);
         }
    }
}

///////////////////////////////////////////////////////////////////////////////
// ルビーインタプリタ初期化
///////////////////////////////////////////////////////////////////////////////
static void myruby_init()
{
    ruby_myinit();

    rb_eval_string("def cursor_move_hook(i) 0; end");
    rb_eval_string("def completion_hook(editing, editing_dir, editing_file, editing_before, earray, cmd, editing_position, squote, dquote, last_squote, last_dquote) return 0; end");
    rb_eval_string("def atexit_hook() 0; end");
    rb_eval_string("def mloop_hook() 0; end");
    rb_eval_string("def mask_hook_ldir(fname, kind) nil; end");
    rb_eval_string("def mask_hook_rdir(fname, kind) nil; end");
    rb_eval_string("def disk_read_hook_ldir() nil; end");
    rb_eval_string("def disk_read_hook_rdir() nil; end");
    rb_eval_string("def path_change_hook(is_ldir) nil; end");
    rb_eval_string("def view_hook() end");
    rb_eval_string("def view_hook_filer() 0 end");

    char tmp[256];
    sprintf(tmp, "$:.push(\"%s\")", SYSCONFDIR);
    rb_eval_string(tmp);
}

///////////////////////////////////////////////////////////////////////////////
// ランタイムスクリプトファイル読み込み
///////////////////////////////////////////////////////////////////////////////
static void read_rc_file()
{
    /// $PREFIX/etc/api.rbを読み込む ///
    char ms_fname[256];
    sprintf(ms_fname, "%s/api.rb", SYSCONFDIR);

    if(access(ms_fname, R_OK) == 0) {
        rb_load(rb_str_new2(ms_fname), 0);
    }
    else {
        fprintf(stderr, "can't find %s/api.rb", SYSCONFDIR);
        exit(1);
    }

    /// $PREFIX/etc/completion.rbを読み込む ///
    char mc_fname[256];
    sprintf(mc_fname, "%s/completion.rb", SYSCONFDIR);

    if(access(mc_fname, R_OK) == 0) {
        rb_load(rb_str_new2(mc_fname), 0);
    }
    else {
        fprintf(stderr, "can't find %s/completion.rb", SYSCONFDIR);
        exit(1);
    }

    /* ~/.mfilerを読み込む。もし ~/.mfiler がないなら$PREFIX/etc/mfiler.rbを読み込む */
    char* home = getenv("HOME");
    if(home == NULL) {
        fprintf(stderr, "$HOME is NULL");
        exit(1);
    }

    char rc_fname[256];
    sprintf(rc_fname, "%s/.mfiler", home);
    
    if(access(rc_fname, R_OK) == 0) {
        rb_load(rb_str_new2(rc_fname), 0);
    }
    else {
        sprintf(rc_fname,"%s/mfiler.rb", SYSCONFDIR);
        if(access(rc_fname, R_OK) == 0) {
            rb_load(rb_str_new2(rc_fname), 0);
        }
        else {
            fprintf(stderr, "can't find $HOME/.mfiler or %s/mfiler.rb file\n", SYSCONFDIR);
            exit(1);
        }
    }
}

///////////////////////////////////////////////////////////////////////////////
// キーボード入力処理
///////////////////////////////////////////////////////////////////////////////
static void input(int meta, int key)
{
    /// ヘルプ ///
    if(gViewHelp) {
        help_input(meta, key);
    }
    /// ディレクトリ選択 ///
    else if(gSelectDir) {
        ActiveDir()->SelectDirInput(meta, key);
    }
    /// コマンドライン ///
    else if(gCmdLineActive) {
        /// ヒストリー選択 ///
        if(vector_size(gHCandidate) > 0)
            cmdline_history_input(meta, key);
        /// 補完候補選択 ///
        else if(gCompletionMenu)
            cmdline_completion_menu_input(meta, key);
        /// 普通のコマンドライン ///
        else
            cmdline_input(meta, key);
    }
    /// メニュー選択 ///
    else if(gActiveMenu) {
        gActiveMenu->Input(meta, key);
    }
    /// インクリメンタルサーチ ///
    else if(gISearch) {
        isearch_input(meta, key);
    }
    /// エキスプローラー風インクリメンタルサーチ ///
    else if(gISearchExplore && IsISearchExploreChar(meta, key)) {
        isearch_explore_input(meta, key);
    }
    /// エキスプローラー風インクリメンタルサーチ 上下キーとスペースキー ///
    else if(gISearchExplore && !IsISearchNULL() 
        && (meta == 0 && key == ' '
            || meta == 1 && (key == 14 || key == 16)
            || meta == 1 && (key == KEY_UP || key == KEY_DOWN)
            || meta == 0 && (key == KEY_F(15) || key == KEY_F(16))
            ))
    {
        isearch_explore_input(meta, key);
    }
    /// ファイラ画面 ///
    else {
        sFile* cursor = ActiveDir()->CursorFile();

        /// ファイル名を得る ///
        const char* name = cursor->mName;
        char fname[1024];
        
        strcpy(fname, name);
        if(gExtensionICase && is_all_ascii((char*)name)) {
            strtolower(fname);
        }

        /// 拡張子名を得る ///
        char extension[256];

        char* tmp = extname(cursor->mName);
        if(gExtensionICase) {
            strtolower(tmp);
        }
        sprintf(extension, ".%s", tmp);
        free(tmp);
    
        /// マーク中ならその状態を調べる ///
        bool is_extension_marked = false;           // 同一拡張子のファイルをマーク中
        char marked_extension[1024];                // マーク中の拡張子
        bool is_directory_marked = false;           // ディレクトリをマーク中
        bool is_executive_file_marked = false;      // 実行ファイルをマーク中
        bool is_link_marked = false;                // リンクをマーク中

        if(ActiveDir()->Marking()) {
            char tmp[1024];
            is_extension_marked = ActiveDir()->IsSameExtensionMarked(tmp);
            if(is_extension_marked) {
                if(gExtensionICase) {   // 大文字小文字を区別しない場合
                    strtolower(tmp);
                }
                sprintf(marked_extension, ".mark-%s", tmp);
            }
            
            is_directory_marked = ActiveDir()->IsDirectoryMarked();
            is_executive_file_marked = ActiveDir()->IsExecutiveFileMarked();
            is_link_marked = ActiveDir()->IsLinkMarked();
        }

        /// 同一拡張子のファイルをマーク中 ///
        if(is_extension_marked
                && key >= 0 && key < KEY_MAX
                && hash_item(gKeyCommand[meta][key], marked_extension))
        {
            int error;
            rb_eval_string_protect((char*)hash_item(gKeyCommand[meta][key],  marked_extension), &error);
            
            output_eval_error(error);

            ISearchClear();
        }

        /// リンクをマーク中 ///
        else if(is_link_marked
                && key >= 0 && key < KEY_MAX
                && hash_item(gKeyCommand[meta][key], ".mark-link"))
        {
            int error;
            rb_eval_string_protect((char*)hash_item(gKeyCommand[meta][key],  ".mark-link"), &error);
            
            output_eval_error(error);

            ISearchClear();
        }

        /// ディレクトリをマーク中 ///
        else if(is_directory_marked
                && key >= 0 && key < KEY_MAX
                && hash_item(gKeyCommand[meta][key], ".mark-dir"))
        {
            int error;
            rb_eval_string_protect((char*)hash_item(gKeyCommand[meta][key],  ".mark-dir"), &error);
            
            output_eval_error(error);

            ISearchClear();
        }

        /// 実行ファイルをマーク中 ///
        else if(is_executive_file_marked
                && key >= 0 && key < KEY_MAX
                && hash_item(gKeyCommand[meta][key], ".mark-execute"))
        {
            int error;
            rb_eval_string_protect((char*)hash_item(gKeyCommand[meta][key],  ".mark-execute"), &error);
            
            output_eval_error(error);

            ISearchClear();
        }

        /// マーク中 ///
        else if(ActiveDir()->Marking()
            && key >= 0 && key < KEY_MAX 
            && hash_item(gKeyCommand[meta][key], ".mark"))
        {
            int error;
            rb_eval_string_protect((char*)hash_item(gKeyCommand[meta][key],".mark"), &error);
            
            output_eval_error(error);

            ISearchClear();
        }

        /// リンクを選択中 ///
        else if(S_ISLNK(cursor->mLStat.st_mode)
            && key >= 0 && key < KEY_MAX 
            && hash_item(gKeyCommand[meta][key], ".link"))
        {
            int error;
            rb_eval_string_protect((char*)hash_item(gKeyCommand[meta][key], ".link"), &error);
            
            output_eval_error(error);

            ISearchClear();
        }

        /// ファイル名で指定されている場合 ///
        else if(key >= 0 && key < KEY_MAX  
              && hash_item(gKeyCommand[meta][key], fname))
        {
            int error;
           rb_eval_string_protect(
               (char*)hash_item(gKeyCommand[meta][key], fname), &error);
            
            output_eval_error(error);

            ISearchClear();
        }

        /// 拡張子名で指定されている場合 ///
        else if(key >= 0 && key < KEY_MAX 
               && hash_item(gKeyCommand[meta][key], extension))
        {
            int error;
            rb_eval_string_protect(
                (char*)hash_item(gKeyCommand[meta][key], extension), &error);
            
            output_eval_error(error);

            ISearchClear();
        }

        /// ディレクトリを選択中 ///
        else if(S_ISDIR(cursor->mLStat.st_mode)
            && key >= 0 && key < KEY_MAX 
            && hash_item(gKeyCommand[meta][key], ".dir"))
        {
            int error;
            rb_eval_string_protect((char*)hash_item(gKeyCommand[meta][key], ".dir"), &error);
            
            output_eval_error(error);

            ISearchClear();
        }

        /// 実行ファイルを選択中 ///
        else if((cursor->mStat.st_mode&S_IXUSR
             || cursor->mStat.st_mode&S_IXGRP || cursor->mStat.st_mode&S_IXOTH)
            && key >= 0 && key < KEY_MAX 
            && hash_item(gKeyCommand[meta][key], ".execute"))
        {
            int error;
            rb_eval_string_protect((char*)hash_item(gKeyCommand[meta][key],".execute"), &error);
            
            output_eval_error(error);

            ISearchClear();
        }

        /// その他のファイル ///
        else if(key >= 0 && key < KEY_MAX 
            && hash_item(gKeyCommand[meta][key], "*")) 
        {
            int error;
            rb_eval_string_protect((char*)hash_item(gKeyCommand[meta][key], "*"), &error);
            
            output_eval_error(error);

            ISearchClear();
        }
        else {
            ISearchClear();
        }
    }

    
}

///////////////////////////////////////////////////////////////////////////////
// 描写
///////////////////////////////////////////////////////////////////////////////

/// ファイラ画面入力時のコマンドライン部の描写 ///
static void view_cmdline_filer()
{
    const int maxx = mgetmaxx();        // 画面サイズ
    const int maxy = mgetmaxy();
    
    /// インクリメンタルサーチ描写 ///
    if(gISearch) {
        isearch_view();
    }
    /// 状態を描写 ///
    else {
        /// 状態を描写 ///
        mmvprintw(maxy -2, 0, "the Minnu's Filer2 %s %s %s %s@%s ", gVersion, cDirWnd::kSortName[ActiveDir()->mSortKind], gKanjiCodeString[gKanjiCodeFileName], gUserName, gHostName);

        /// 他端末実行機能を使っている場合xを描写 ///            
        if((gXterm && !(gXtermNext||gXterm2)) || (!gXterm && (gXtermNext||gXterm2))) mprintw("x ");

        if(gISearchExplore) isearch_explore_view();
    }


    /// カーソル下のファイルの情報を描写 ///          
    char buf[1024];
    sFile* file = ActiveDir()->CursorFile();

    char permission[12];
    strcpy(permission, "----------");

    if(S_ISDIR(file->mLStat.st_mode)) permission[0] = 'd';  // ファイルの種別
    if(S_ISCHR(file->mLStat.st_mode)) permission[0] = 'c';
    if(S_ISBLK(file->mLStat.st_mode)) permission[0] = 'b';
    if(S_ISFIFO(file->mLStat.st_mode)) permission[0] = 'p';
    if(S_ISLNK(file->mLStat.st_mode)) permission[0] = 'l';
    if(S_ISSOCK(file->mLStat.st_mode)) permission[0] = 's';

    mode_t smode = file->mLStat.st_mode;        // パーミッション
    if(smode & S_IRUSR) permission[1] = 'r';
    if(smode & S_IWUSR) permission[2] = 'w';
    if(smode & S_IXUSR) permission[3] = 'x';
    if(smode & S_IRGRP) permission[4] = 'r';
    if(smode & S_IWGRP) permission[5] = 'w';
    if(smode & S_IXGRP) permission[6] = 'x';
    if(smode & S_IROTH) permission[7] = 'r';
    if(smode & S_IWOTH) permission[8] = 'w';
    if(smode & S_IXOTH) permission[9] = 'x';
    if(smode & S_ISUID) permission[3] = 's';
    if(smode & S_ISGID) permission[6] = 's';
#if defined(S_ISTXT)
    if(smode & S_ISTXT) permission[9] = 't';
#endif
#if defined(S_ISVTX)
    if(smode & S_ISVTX) permission[9] = 't';
#endif
    
    permission[10] = 0;

    time_t t = file->mLStat.st_mtime;                   // 時間
    struct tm* tm_ = (struct tm*)localtime(&t);
    
    char buf2[1024];

    char owner[256];
    char* tmp = file->mUser;
    if(tmp) {
        if(gKanjiCode == kUtf8)
            str_cut2(tmp, 8, owner, 256);
        else 
            cut(tmp, owner, 8);
    }
    else {
        sprintf(owner, "%d", file->mLStat.st_uid);
    }

    char group[256];
    tmp = file->mGroup;
    if(tmp) {
        if(gKanjiCode == kUtf8)
            str_cut2(tmp, 7, group, 256);
        else
            cut(tmp, group, 7);
    }
    else {
        sprintf(group, "%d", file->mLStat.st_gid);
    }

    if(!S_ISLNK(file->mLStat.st_mode)) {
        int year = tm_->tm_year-100;
        if(year < 0) year+=100;
        while(year > 100) year-=100;
        
        sprintf(buf2
            , "%s %3d %-8s %-7s%10lld %02d-%02d-%02d %02d:%02d %s"
            , permission, file->mLStat.st_nlink
            , owner, group
                                
            , file->mLStat.st_size
            
            , year, tm_->tm_mon+1, tm_->tm_mday
            , tm_->tm_hour, tm_->tm_min
            , file->mNameView);
            
        char buf3[1024];
        if(gKanjiCode == kUtf8) 
            str_cut2(buf2, maxx-1, buf3, 1024);
        else
            cut(buf2, buf3, maxx-1);
                            
        mmvprintw(maxy-1, 0, "%s", buf3);
    }
    else {
        sprintf(buf
           , "%s %3d %s%s%10lld %02d-%02d-%02d %02d:%02d %s -> %s"
           , permission, file->mLStat.st_nlink
           , owner, group
                                
           , file->mLStat.st_size

           , tm_->tm_year-100, tm_->tm_mon+1, tm_->tm_mday
           , tm_->tm_hour, tm_->tm_min
           , file->mNameView
           , file->mLinkTo);
           
        char buf2[1024];
        if(gKanjiCode == kUtf8)
            str_cut2(buf, maxx-1, buf2, 1024);
        else
            cut(buf, buf2, maxx-1);
                   
        mmvprintw(maxy-1, 0, "%s", buf2);
    }

    mmove(maxy-1, maxx-2);
}

void view(bool cmdline)
{
    /// ヘルプ画面 ///
    if(gViewHelp) {
        help_view();
    }

    /// メニュー画面 ///
    else if(gActiveMenu) {
        /// ファイラ画面 ///
        gLDir->View();
        gRDir->View();

        /// ジョブを描写 ///
        shell_view();

        /// コマンドラインを描写 ///
        if(cmdline) {
            view_cmdline_filer();
        }

        /// メニューを描写 ///
        gActiveMenu->View();
    }

    /// ディレクトリ選択画面 ///
    else if(gSelectDir) {
        ActiveDir()->SelectDirView();
    }

    /// コマンドライン入力時 ///
    else if(gCmdLineActive) {
        /// ヒストリー選択画面 ///
        if(vector_size(gHCandidate))
            cmdline_history_view();
        /// 補完候補選択画面 ///
        else if(vector_size(gCCandidate))
            cmdline_completion_view();
        /// ファイラ画面 ///
        else {
            gLDir->View();
            gRDir->View();

            /// ジョブを描写 ///
            shell_view();
        }

        /// コマンドラインの描写 ///
        if(cmdline) {
            cmdline_view();
        }
    }

    /// ファイラ画面 ///
    else {
        /// ディレクトリの描写 ///
        gLDir->View();
        gRDir->View();

        /// ジョブを描写 ///
        shell_view();

        /// コマンドラインの描写 ///
        if(cmdline) {
            view_cmdline_filer();
        }
    }

    /// 登録描写関数 ///
    if(gViewSystem) gViewSystem();
    if(gView) {
        gView();
    }
    else {
        rb_funcall(rb_cObject, rb_intern("view_hook"), 0);
    }
}

///////////////////////////////////////////////////////////////////////////////
// exit時に実行される関数
///////////////////////////////////////////////////////////////////////////////
void atexit_fun()
{
    if(mis_curses()) {
        mendwin();
    }
}

///////////////////////////////////////////////////////////////////////////////
// シグナル処理
///////////////////////////////////////////////////////////////////////////////
static void sig_winch(int signal)
{
    if(gMainLoop) {
        mclear_immediately();       // 画面の再描写
        view(true);
        mrefresh();
    }
}

void set_signal_clear()
{
    signal(SIGABRT, SIG_DFL);
    signal(SIGALRM, SIG_DFL);
    signal(SIGBUS, SIG_DFL);
    signal(SIGCHLD, SIG_DFL);
    signal(SIGCONT, SIG_DFL);
    signal(SIGHUP, SIG_DFL);
    signal(SIGFPE, SIG_DFL);
    signal(SIGILL, SIG_DFL);
    signal(SIGINT, SIG_DFL);
    signal(SIGIO, SIG_DFL);
    signal(SIGKILL, SIG_DFL);
    signal(SIGQUIT, SIG_DFL);
    signal(SIGPIPE, SIG_DFL);
    signal(SIGPROF, SIG_DFL);
    signal(SIGSEGV, SIG_DFL);
    signal(SIGSTOP, SIG_DFL);
    signal(SIGTERM, SIG_DFL);
    signal(SIGTRAP, SIG_DFL);
    signal(SIGTSTP, SIG_DFL);
    signal(SIGTTIN, SIG_DFL);
    signal(SIGTTOU, SIG_DFL);
    signal(SIGWINCH, SIG_DFL);
    signal(SIGURG, SIG_DFL);
    signal(SIGUSR1, SIG_DFL);
    signal(SIGUSR2, SIG_DFL);
    signal(SIGXCPU, SIG_DFL);
    signal(SIGXFSZ, SIG_DFL);
    signal(SIGVTALRM, SIG_DFL);
}

void set_signal_mfiler2()
{
    signal(SIGWINCH, sig_winch);
    signal(SIGTTOU, SIG_IGN);
    signal(SIGTTIN, SIG_IGN);
    signal(SIGINT, SIG_IGN);
    signal(SIGQUIT, SIG_IGN);
    signal(SIGTSTP, SIG_IGN);
    signal(SIGCHLD, SIG_IGN);
}

///////////////////////////////////////////////////////////////////////////////
// メイン関数
///////////////////////////////////////////////////////////////////////////////
int main(int argc, char* argv[])
{
CHECKML_BEGIN();
LOG_BEGIN();

    signal(SIGWINCH, SIG_IGN);

    /// 標準入力と出力が制御端末かどうか確認 ///
    if(!isatty(0) || !isatty(1)) {
        fprintf(stderr, "standard input is not a tty\n");
        return 1;
    }

    /// オプション処理 ///
    bool refresh = false;
    bool move_dir = false;
    char move_dir_path[PATH_MAX];
    
    char init_path[PATH_MAX];   // 初期ディレクトリ
    strcpy(init_path, "");
    
    int opt;
    while((opt = getopt(argc, argv, "rm:i:v")) != -1) {
        switch(opt) {
        case 'r':
            refresh = true;
            break;

        case 'm':
            move_dir = true;
            sprintf(move_dir_path, "m%s", optarg);
            break;

        case 'i':
            strcpy(init_path, optarg);
            break;

        case 'v':
            printf("the Minnu's Filer2 version %s by Daisuke Minato\n", gVersion);
            exit(0);
            break;

        case ':':
            printf("option needs a value\n");
            exit(1);
            break;

        case '?':
            printf("usage mfiler2 [-r | -m path | -i path]\n\n");
            printf("-r : send a command to running all mfiler2. The command is rereading disk and refreshing screen\n");
            printf("-m path : send a command to running all mfiler2. The command is moving to argument path\n");
            printf("-i path : set init path\n");
            exit(1);
            break;
        }
    }


    /// オプションで指定されているなら動いているmfiler2に画面リフレッシュのメッセージを送る  ///
    if(refresh && gUnixDomainSocket) {
        DIR* dir = opendir("/tmp");
        if(dir == NULL) {
            fprintf(stderr, "/tmp doesn't exist");
            exit(1);
        }

        struct dirent* entry;
        while(entry = readdir(dir)) {
            if(strstr(entry->d_name, "mfiler")) {
                char soc_name[256];
                sprintf(soc_name, "/tmp/%s", entry->d_name);

                int soc = socket(AF_UNIX, SOCK_DGRAM, 0);
                if(soc < 0) {
                    perror("socket");
                    return 1;
                }

                struct sockaddr_un addr;
                addr.sun_family = AF_UNIX;
                strcpy(addr.sun_path, soc_name);
                if(connect(soc, (struct sockaddr *)&addr, sizeof(addr)) == 0) {
                    write(soc, "r", 2);
                }

                close(soc);
            }
        }
        
        return 0;
    }
    if(move_dir && gUnixDomainSocket) {
        DIR* dir = opendir("/tmp");
        if(dir == NULL) {
            fprintf(stderr, "/tmp doesn't exist");
            exit(1);
        }

        struct dirent* entry;
        while(entry = readdir(dir)) {
            if(strstr(entry->d_name, "mfiler")) {
                char soc_name[256];
                sprintf(soc_name, "/tmp/%s", entry->d_name);

                int soc = socket(AF_UNIX, SOCK_DGRAM, 0);
                if(soc < 0) {
                    perror("socket");
                    return 1;
                }

                struct sockaddr_un addr;
                addr.sun_family = AF_UNIX;
                strcpy(addr.sun_path, soc_name);
                if(connect(soc, (struct sockaddr *)&addr, sizeof(addr)) == 0) {
                    write(soc, move_dir_path, strlen(move_dir_path)+1);
                }

                close(soc);
            }
        }
        
        return 0;
    }

    /// フォワグランドジョブになるまでループ ///
    int status;
    pid_t pgrp;
    while((status = tcgetpgrp(0)) >= 0) {
        if(status == (pgrp = getpgrp())) {
            break;
        }
        kill(- pgrp, SIGTTIN);
    }

    /// 新しいセッションを始めてプロセスグループリーダーになる ///
    setsid();
    pid_t shell_pgrp = getpid();
    setpgid(shell_pgrp, shell_pgrp);

    /// 端末の制御を握る ///
    tcsetpgrp(0, shell_pgrp);

    /// Rubyインタプリタ初期化 ///
    myruby_init();
    
    /// モジュールの初期化 ///
    command_init();
    cMenu::Init();
    cDirWnd::Init();
    mcurses_init();
    cmdline_init();

    /// グローバル変数の初期化 ///
    initialize_global_vars(init_path);

    /// ソケットの初期化 ///
    char soc_name[256];
    int soc;
    if(gUnixDomainSocket) {
        sprintf(soc_name, "/tmp/mfiler%d", getpid());

        if(access(soc_name, F_OK) == 0) {
            unlink(soc_name);
        }
        
        soc = socket(AF_UNIX, SOCK_DGRAM, 0);
        if(soc < 0) {
            perror("socket");
            return 1;
        }
        
        struct sockaddr_un addr;
        addr.sun_family = AF_UNIX;
        strcpy(addr.sun_path, soc_name);
        if(bind(soc, (struct sockaddr *)&addr, sizeof(addr)) == -1) {
            perror("bind");
            return 1;
        }
    }

    /// ランタイムスクリプト実行 ///
    read_rc_file();

    cmdline_init_after_reading_rc_file();
    isearch_init_after_reading_rc_file();

    /// localeの設定 ///
    if(gKanjiCode == kUtf8) setlocale(LC_CTYPE, "ja_JP.UTF-8");

    /// オプションによる初期ディレクトリの指定 ///
    if(strcmp(init_path, "") != 0) {
        gLDir->Move(init_path);
        gLDir->SetScrollTop(0);
        gLDir->MoveCursor(0);

        gRDir->Move(init_path);
        gRDir->SetScrollTop(0);
        gRDir->MoveCursor(0);
    }

    /// アクティブディレクトリを設定 ///
    gLDir->Activate(gRDir);

    /// シグナルを設定 ///
    set_signal_mfiler2();
    
    /// カーシスの起動 ///
    minitscr();
    mclear_immediately();

    /// atexit登録 ///
    atexit(atexit_fun);
    
    /// メインループ ///
    gMainLoop = true;

    fd_set mask;
    fd_set read_ok;

    FD_ZERO(&mask);
    FD_SET(0, &mask);
    if(gUnixDomainSocket) FD_SET(soc, &mask);

    while(gMainLoop) {
        /// メインループフック ///
        rb_funcall(rb_cObject, rb_intern("mloop_hook"), 0);
        
        /// 描写 ///
        mclear();
        view(true);
        mrefresh();

        /// selectで入力待ち ///
        read_ok = mask;

        if(!mkbuf_exist()) {            // キーボードのバッファがないならselect
            if(gUnixDomainSocket) {
                if(vector_size(gJobs) > 0) {
                    struct timeval tv;
                    tv.tv_sec = 5;
                    tv.tv_usec = 0;

                    select(soc + 1, &read_ok, NULL, NULL, &tv);
                }
                else {
                    select(soc + 1, &read_ok, NULL, NULL, NULL);
                }
            }
            else {
                if(vector_size(gJobs) > 0) {
                    struct timeval tv;
                    tv.tv_sec = 5;
                    tv.tv_usec = 0;

                    select(1, &read_ok, NULL, NULL, &tv);
                }
                else {
                    select(1, &read_ok, NULL, NULL, NULL);
                }
            }
        }

        /// キー処理 ///
        if(FD_ISSET(0, &read_ok) || mkbuf_exist()) {    // キー入力があったか、キーボードのバッファがあるなら
            /// コマンドライン入力中 ///
            if(gCmdLineActive) {
                while(1) {
                    /// 漢字入力のためselectで入力を見張る ///
                    struct timeval tv;
                    tv.tv_sec = 0;
                    tv.tv_usec = 0;

                    read_ok = mask;

                    select(1, &read_ok, NULL, NULL, &tv);

                    /// キー入力中 ///
                    if(FD_ISSET(0, &read_ok) || mkbuf_exist()) {
                        int meta;
                        int key = mgetch(&meta);

                        if(key == -1) {
                        }
                        else if(meta) {
                            input(1, key);
                        }
                        /// normal key ///
                        else {
                           input(0, key);
                        }
                    }
                    /// キー入力が終わった ///
                    else {
                        break;
                    }
                }
            }

            /// その他 ///
            else {
                int meta;
                int key = mgetch_nonblock(&meta);

                /// ESCキー ///
                if(key == -1) {
                }
                else if(meta) {
                    input(1, key);
                }
                
                /// meta key ///
                else if(key >= kKeyMetaFirst && key <= kKeyMetaFirst+127) {
                    input(1, key - kKeyMetaFirst);
                }

                /// normal key ///
                else {
                   input(0, key);
                }
            }
        }

        /// UNIX Domainソケット処理 ///
        else if(gUnixDomainSocket && FD_ISSET(soc, &read_ok)) {
            char buf[PATH_MAX];
            
            struct sockaddr_un caddr;
            socklen_t caddr_len;
            int len = recvfrom(soc, buf, PATH_MAX, 0, (struct sockaddr *)&caddr, &caddr_len);

            if(buf[0] == 'r') {
                gLDir->Reread();
                gRDir->Reread();
            }
            else if(buf[0] == 'm') {
                ActiveDir()->Move(buf + 1);
            }
        }

        /// ジョブ処理 ///
        shell_wait_backgroud_job();
    }

    /// カーシス終わり ///
    mmove_immediately(0,0);
    mclear_immediately();
    mendwin();

    /// シグナルを初期化 ///
    set_signal_clear();
    
    /// 全てのジョブをkill ///
    shell_kill_all_jobs();
    
    /// 終了フック ///
    rb_funcall(rb_cObject, rb_intern("atexit_hook"), 0);
    ruby_myfinal();
    putp(tigetstr("rmkx"));

    /// モジュール解放 ///
    cDirWnd::Final();
    cmdline_final();
    command_final();
    cMenu::Final();
    mcurses_final();
    isearch_final();
    
    /// グローバル変数解放 ///    
    finalize_global_vars();

    /// ソケット解放 ///
    if(gUnixDomainSocket) unlink(soc_name);

CHECKML_BEGIN();
LOG_BEGIN();

    return 0;
}

